# Copyright 2018 The Bazel Authors. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Partial implementation for debug symbol file processing."""

load(
    "@build_bazel_rules_apple//apple/internal:bundling_support.bzl",
    "bundling_support",
)
load(
    "@build_bazel_rules_apple//apple/internal:file_support.bzl",
    "file_support",
)
load(
    "@build_bazel_rules_apple//apple/internal/utils:defines.bzl",
    "defines",
)
load(
    "@bazel_skylib//lib:partial.bzl",
    "partial",
)
load(
    "@bazel_skylib//lib:paths.bzl",
    "paths",
)

# TODO(b/110264170): Expose this provider so that IDEs can use it to reference the dSYM bundles
# contained in the dsym_bundles field.
_AppleDebugInfo = provider(
    doc = """
Private provider to propagate transitive debug symbol information.
""",
    fields = {
        "dsym_bundles": """
Paths to dSYM bundles that this target provides. This includes the paths to dSYM bundles generated
for dependencies of this target (e.g. frameworks and extensions).
""",
        "dsyms": """
Depset of `File` references to dSYM files if requested in the build with --apple_generate_dsym.
""",
        "linkmaps": """
Depset of `File` references to linkmap files if requested in the build with --objc_generate_linkmap.
""",
    },
)

def _collect_linkmaps(ctx, debug_provider, bundle_name):
    """Collects the available linkmaps from the binary.

    Args:
      ctx: The target's rule context.
      debug_provider: The AppleDebugOutput provider for the binary target.
      bundle_name: The name of the output bundle.

    Returns:
      A list of linkmap files, one per linked architecture.
    """
    outputs = []

    # TODO(b/36174487): Iterate over .items() once the Map/dict problem is fixed.
    for arch in debug_provider.outputs_map:
        arch_outputs = debug_provider.outputs_map[arch]
        linkmap = arch_outputs["linkmap"]
        output_linkmap = ctx.actions.declare_file(
            "%s_%s.linkmap" % (bundle_name, arch),
        )
        outputs.append(output_linkmap)
        file_support.symlink(ctx, linkmap, output_linkmap)

    return outputs

# TODO(kaipi): Investigate moving this actions into a tool so that we can use
# tree artifacts instead, which should simplify parts of this file.
def _bundle_dsym_files(ctx, debug_provider, bundle_name, bundle_extension = ""):
    """Recreates the .dSYM bundle from the AppleDebugOutputs provider.

    The generated bundle will have the same name as the bundle being built (including its
    extension), but with the ".dSYM" extension appended to it.

    If the target being built does not have a binary or if the build it not generating debug
    symbols (`--apple_generate_dsym` is not provided), then this function is a no-op that returns
    an empty list.

    Args:
      ctx: The target's rule context.
      debug_provider: The AppleDebugOutput provider for the binary target.
      bundle_name: The name of the output bundle.
      bundle_extension: The extension for the bundle.

    Returns:
      A list of files that comprise the .dSYM bundle, which should be returned as additional
      outputs from the target.
    """
    bundle_name_with_extension = bundle_name + bundle_extension
    dsym_bundle_name = bundle_name_with_extension + ".dSYM"

    outputs = []

    # TODO(b/36174487): Iterate over .items() once the Map/dict problem is fixed.
    for arch in debug_provider.outputs_map:
        arch_outputs = debug_provider.outputs_map[arch]
        dsym_binary = arch_outputs["dsym_binary"]
        output_binary = ctx.actions.declare_file(
            "%s/Contents/Resources/DWARF/%s_%s" % (
                dsym_bundle_name,
                bundle_name,
                arch,
            ),
        )
        outputs.append(output_binary)
        file_support.symlink(ctx, dsym_binary, output_binary)

    # If we found any outputs, create the Info.plist for the bundle as well; otherwise, we just
    # return the empty list. The plist generated by dsymutil only varies based on the bundle name,
    # so we regenerate it here rather than propagate the other one from the apple_binary. (See
    # https://github.com/llvm-mirror/llvm/blob/master/tools/dsymutil/dsymutil.cpp)
    if outputs:
        dsym_plist = ctx.actions.declare_file(
            "%s/Contents/Info.plist" % dsym_bundle_name,
        )
        outputs.append(dsym_plist)
        ctx.actions.expand_template(
            output = dsym_plist,
            template = ctx.file._dsym_info_plist_template,
            substitutions = {
                "%bundle_name_with_extension%": bundle_name_with_extension,
            },
        )

    return outputs

def _debug_symbols_partial_impl(ctx, debug_dependencies = [], debug_outputs_provider = None):
    """Implementation for the debug symbols processing partial."""
    deps_providers = [
        x[_AppleDebugInfo]
        for x in debug_dependencies
        if _AppleDebugInfo in x
    ]

    dsym_bundles = depset(transitive = [x.dsym_bundles for x in deps_providers])

    direct_dsyms = []
    transitive_dsyms = [x.dsyms for x in deps_providers]

    direct_linkmaps = []
    transitive_linkmaps = [x.linkmaps for x in deps_providers]

    output_providers = []

    if debug_outputs_provider:
        output_providers.append(debug_outputs_provider)

        bundle_name = bundling_support.bundle_name(ctx)

        if ctx.fragments.objc.generate_dsym:
            bundle_extension = bundling_support.bundle_extension(ctx)
            dsym_files = _bundle_dsym_files(
                ctx,
                debug_outputs_provider,
                bundle_name,
                bundle_extension,
            )
            direct_dsyms.extend(dsym_files)

            absolute_dsym_bundle_path = paths.join(
                ctx.bin_dir.path,
                ctx.label.package,
                bundle_name + bundle_extension + ".dSYM",
            )
            dsym_bundles = depset(
                [absolute_dsym_bundle_path],
                transitive = [dsym_bundles],
            )

        if ctx.fragments.objc.generate_linkmap:
            linkmaps = _collect_linkmaps(ctx, debug_outputs_provider, bundle_name)
            direct_linkmaps.extend(linkmaps)

    # Only output dependency debug files if requested.
    # TODO(b/131699846): Remove this.
    propagate_embedded_extra_outputs = defines.bool_value(
        ctx,
        "apple.propagate_embedded_extra_outputs",
        False,
    )

    dsyms_group = depset(direct_dsyms, transitive = transitive_dsyms)
    linkmaps_group = depset(direct_linkmaps, transitive = transitive_linkmaps)

    if propagate_embedded_extra_outputs:
        output_files = depset(transitive = [dsyms_group, linkmaps_group])
    else:
        output_files = depset(direct_dsyms + direct_linkmaps)

    output_providers.append(
        _AppleDebugInfo(
            dsym_bundles = dsym_bundles,
            dsyms = dsyms_group,
            linkmaps = linkmaps_group,
        ),
    )

    return struct(
        output_files = output_files,
        providers = output_providers,
        output_groups = {
            "dsyms": dsyms_group,
            "linkmaps": linkmaps_group,
        },
    )

def debug_symbols_partial(debug_dependencies = [], debug_outputs_provider = None):
    """Constructor for the debug symbols processing partial.

    This partial collects all of the transitive debug files information. The output of this partial
    are the debug output files for the target being processed _plus_ all of the dependencies debug
    symbol files. This includes dSYM bundles and linkmaps. With this, for example, by building an
    ios_application target with --apple_generate_dsym, this partial will return the dSYM bundle of
    the ios_application itself plus the dSYM bundles of any ios_framework and ios_extension
    dependencies there may be, which will force bazel to present these files in the output files
    section of a successful build.

    Args:
      debug_dependencies: List of targets from which to collect the transitive dependency debug
        information to propagate them upstream.
      debug_outputs_provider: The AppleDebugOutputs provider containing the references to the debug
        outputs of this target's binary.

    Returns:
      A partial that returns the debug output files, if any were requested.
    """
    return partial.make(
        _debug_symbols_partial_impl,
        debug_dependencies = debug_dependencies,
        debug_outputs_provider = debug_outputs_provider,
    )
